/**************************************************************************************
 * Filename: Com_vc.cpp
 * Author  : Michael D. Hall
 * Date    : 4/23/2002
 * 
 * Description
 * This collection of C routines is intended to demonstrate bi-directional 
 * communication of Boca Ticket Printers.  It is possible to read and write from
 * the serial port using the routines named "Port_Read" and "Port_Write". 
 *
 * This is but one method that works.  There are many other possible ways, which I
 * did not explore.
 *
 * The code has functions designed for both types of Boca printers, the 
 * FGL & HP emulation models.  Print FGL test tickets on FGL printers and HP test tickets
 * on HP emulation printers. If you print the wrong test ticket on your printer,
 * the results will be mixed, but not harmful.  
 *
 * When prompted to enter text, you can enter FGL commands like: 
 *
 *                    <RC200,200><F5>This is a test<p>
 *
 * for an FGL printer or enter simply text to be printed through on an HP emulation model,
 * typically followed by an ASCII Form Feed.
 *
 * When prompted for the decimal ASCII value enter a number from the ASCII table.
 * Some of the more popular HP printer commands are the following ASCII numbers.
 *   CR  = 13       'Use this for a carriage return
 *   LF  = 10       'Use this for a line feed
 *   TAB = 9        'Use this for a horizontal Tab
 *   FF  = 12       'Use this for a form feed
 *   ESC = 27       'Use this for an Escape
 *
 **************************************************************************************/
#include <conio.h> /* required only for function declarations */
#include "stdafx.h"
#include "stdio.h"
#include "stdlib.h"
#include <stdio.h>
#include <windows.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <io.h>
#include "com_vc.h"

/**************************************************************************************
 * Prototype Definitions of local routines
 **************************************************************************************/
int __kbhit( void );

void Display_Buffer(int size);
void Port_Read();
void Port_Write();
void Clear_Buffer();
void fgl_test_ticket();
void hp_test_ticket();
void Write();
void Perform_Test();
void Setup_Port();
void Open_Port(int Port_Num);
void Shutdown_Port();
void Clear_Comm();
void look_for_data(int item);

/**************************************************************************************
 * Global variables
 **************************************************************************************/
	char Buffer[PACKET_SIZE+1];	// char string buffer
	PVOID pBuffer;				// pointer to buffer
	FILE *fin, *fout;			// Input/Output files
	int port;					// Choice of ports as listed below
	char port_name[5];			// 4 char name + 1 for null
	int rw_function;			// read=1 & write=2
	int ascii_value;				// Ascii value for printer
	int w_function;				// FGL Command=1  & Test Ticket=2
	DWORD error_code;			// Error code
	int i,k;
	DWORD dwSize;
	int Bytes_Transferred;
	int pipe;
	HANDLE hWD;
	HINSTANCE hinstLib;
	char ichar;
	HANDLE hPort = NULL;				// Serial Port file pointer
	BOOL     Port_Ready;
	DCB      dcb;
	COMMTIMEOUTS CommTimeouts;
	BOOL     Write_Com;
	BOOL     Read_Com;
	DWORD Bytes_Written;
	DWORD Bytes_Read;
	int ix,start,size;
	int Last_Port_Num = 0;			//Used to track last com port used.
	int default_BaudRate;
	int default_Parity;
	int default_ByteSize;
	int default_StopBits;
	bool found;

/**************************************************************************************
 * Routine Name: main
 * 
 * Description
 * This routine demonstrates a technique for reading and writing from a Boca Systems
 * printer.  The user will be prompted for several options beginning with whether to
 * read from the printer buffer or write to the printer. The user can issue individual
 * commands to the printer or print an FGL test ticket. If FGL commands are stored in
 * file, the code could be modified to open the file, read the commands, send to the
 * printer using the same techniques.
 **************************************************************************************/
int main(int argc, char* argv[])
{

	//look_for_data(0x11);		//Hex 11 is X-ON

      
	printf("Sample program to read/write from/to serial port Boca printers\n");
	printf("==============================================================\n");
	do
	{

		/* Clear Buffer & reset write function between uses */
		Clear_Buffer();
		w_function=1;

		/* Does the user want to read or write across a port? */
		printf("\n\nChoose Read/Write Function\n");
		printf("==========================\n");
		printf(" 1. Read from port\n");
		printf(" 2. Write to port\n");
		printf("99. Exit\n");
		printf("\nChoose function: ");
		scanf("%d",&rw_function);
		
		/* If read/write is anything except 1 or 2 then exit. */
		if(rw_function >= 1 && rw_function <= 2)
		{
			/* Which port does the user want? */
			printf("\n\nChoose Port\n");
			printf("==========================\n");
			printf(" 1. COM1:\n");
			printf(" 2. COM2:\n");
			printf(" 3. COM3:\n");
			printf("99. Exit\n");
			printf("\nEnter Port choice: ");
			scanf("%d",&port);

			if((port>=COM1) && (port<=COM3))	//Perform function only if port number within range
			{

				/* If read/write is write then prompt for what to write. */
				if (rw_function==2)
				{
					printf("\n\nChoose Write Function\n");
					printf("==========================\n");
					printf(" 1. Send Text\n");
					printf(" 2. Send ASCII value\n");
					printf(" 3. Print FGL Test Ticket\n");
					printf(" 4. Print HP Test Ticket\n");
					printf("99. Exit\n");
					printf("\nChoose function: ");
					scanf("%d",&w_function);
					gets(Buffer);			//Terminates last scanf into dummy buffer

					if (w_function==1)
					{
						Clear_Buffer();
						printf("\n\nEnter Text: ");
						gets(Buffer);
					}

					if (w_function==2)
					{
						Clear_Buffer();
						printf("\n\nEnter Decimal ASCII Value: ");
						scanf("%d",&ascii_value);
						Buffer[0] = ascii_value;
					}
					if ((w_function<1) || (w_function>4)) rw_function=0; //exit
				}

				switch (rw_function)		//Are we to perform a read, write or exit?
				{
				case 1:						//Read from port

					switch (port)			//Open port for input
					{
					case COM1:
						// Open the serial port, if not already open.
						strcpy(port_name,COM1_S);
						if (hPort == NULL) Open_Port(COM1);
					break;
					case COM2:
						// Open the serial port, if not already open.
						strcpy(port_name,COM2_S);
						if (hPort == NULL) Open_Port(COM2);
					break;
					case COM3:
						// Open the serial port, if not already open.
						strcpy(port_name,COM3_S);
						if (hPort == NULL) Open_Port(COM3);
					break;
					default:
						// Port # Error, but should never be able to arrive here
						// due to error checking above.
						printf("\nPort read error port = %d\n",port);
						port=0;
					break;
					}

					// close serial port here
					if (hPort == NULL) 
					{
						printf("Printer not found on port %s\n",port_name);
					}
					else
					{
						Port_Read();
						Shutdown_Port();
					}
				break;

				case 2:							//Write to port
					switch (port)				//Open port for output
					{
					case COM1:
						// Open the serial port.
						strcpy(port_name,COM1_S);
						if (hPort == NULL) Open_Port(COM1);
					break;
					case COM2:
						// Open the serial port.
						strcpy(port_name,COM2_S);
						if (hPort == NULL) Open_Port(COM2);
					break;
					case COM3:
						// Open the serial port.
						strcpy(port_name,COM3_S);
						if (hPort == NULL) Open_Port(COM3);
					break;
					default:
						// Port # Error, but should never be able to arrive here
						// due to error checking above.
						printf("\nPort write error port = %d\n",port);
						port = 0;
					break;
					}

					if (hPort == NULL) 
					{
						printf("Printer not found on port %s\n",port_name);
					}
					else
					{
						Perform_Test();
					}
				
				break;
				default:
					// Used to exit cleanly
					//printf("\nExit cleanly port = %d\n",port);
					port = 0;
				break;
				}
			}
			else port=0;
		}
		else port=0;
	}
	while (port != 0);

	if(hPort) Shutdown_Port();					//If a serial port is open, close it.
	return 0;									//We are done.
}

void look_for_data(int item)
{
	int i;
	found = false;
	hPort = NULL;
	strcpy(port_name,COM1_S);
	if (hPort == NULL) Open_Port(COM1);
	do
	{
		Bytes_Transferred = 0;
		do
		{
			
			Read_Com = ReadFile(hPort, &Buffer, 128, &Bytes_Read, NULL);
			Bytes_Transferred = Bytes_Read;

		} while(Bytes_Transferred == 0);

		for (i = 0; i < Bytes_Transferred; i++)
		{
			printf ("%x ", Buffer[i]);
			if (Buffer[i] == 0x10)
			{
				printf (" %s ", "Out of Stock \n");
			}
			if (Buffer[i] == 0x6)
			{
				printf (" %s ", "ACK \n");
			}
			if (Buffer[i] == 0x11)
			{
				printf (" %s ", "XON \n");
			}
			if (Buffer[i] == 0x12)
			{
				printf (" %s ", "Power ON \n");
			}
			if (Buffer[i] == item) found = true;
		}
	} while(found == false);
	printf ("\n\n");
	//Shutdown_Port();
}
/**************************************************************************************
 * Routine Name: Display_Buffer
 * 
 * Description
 * Display whatever is in the buffer two ways Hexadecimal and characters.  The character
 * response prints only printable text.  The Hexadecimal dump will display everything read.
 *
 * Once it is assumed the incoming data is status data it will represent:
 *
 *                        0x02 = "Start of Text"
 *                        0x03 = "End of Text"
 *                        0x04 = "End of Transmission"
 *                        0x05 = "Test Button Ticket ACK"
 *                        0x06 = "Ticket ACK"
 *                        0x10 = "Out of Tickets"
 *                        0x11 = "X-On"
 *                        0x12 = "Power On"
 *                        0x13 = "X-Off"
 *                        0x15 = "Ticket NAK"
 *                        0x16 = "Ribbon Low"
 *                        0x17 = "Ribbon Out"
 *                        0x18 = "Ticket Jam"
 *                        0x19 = "Illegal Data"
 *                        0x1A = "Power Up Problem"
 *                        0x1C = "Incomplete Logo Stored in FLASH"
 **************************************************************************************/
void Display_Buffer(int size)
{
	int i,start,end;

	start = 0;
	end = size;

	printf ("\n\nThe buffer contains %d Byte(s).",size);

	// Display as Ascii characters
	printf ("\n\nCharacter Response   = ");
	for (i = start; i < end; i++)
	{
		// Display only printable characters. Display a space for the rest.
		if ((Buffer[i] >= SPACE) && (Buffer[i] <= TILDE))
			printf ("%c", Buffer[i]);
		else
			printf("%c",SPACE);
	}

	// Display as Ascii Hexadecimal
	printf ("\n\nHexadecimal Response = ");
	for (i = start; i < end; i++)
		printf ("%x ", Buffer[i]);
	printf (" \n");


	return;
}


/**************************************************************************************
 * Routine Name: Port_Read
 * 
 * Description
 * Read from currently selected serial port and display Buffer.
 **************************************************************************************/
void Port_Read()
{

	Bytes_Transferred = 0;
			
	Read_Com = ReadFile(hPort, &Buffer, 128, &Bytes_Read, NULL);

	//When you are through using the file handle simply close the file.
	Bytes_Transferred = Bytes_Read;

	if (Bytes_Transferred == 0)
		printf ("\nNothing was read from port %s\n",port_name);
	else
		Display_Buffer(Bytes_Transferred);
	return;
}


/**************************************************************************************
 * Routine Name: Port_Write
 * 
 * Description
 * Write Buffer out currently select serial port.
 **************************************************************************************/
void Port_Write()
{

	i = strlen(Buffer);
	if (i > 0)
		Write_Com = WriteFile(hPort, Buffer,i,&Bytes_Written,NULL);

	return;
}
/**************************************************************************************
 * Routine Name: Clear_Buffer()
 * 
 * Description
 * Clear Buffer between uses by padding string with nulls.
 **************************************************************************************/
void Clear_Buffer()
{
	pBuffer = &Buffer;
	memset (pBuffer,0,PACKET_SIZE+1);
	return;
}

/**************************************************************************************
 * Routine Name: fgl_test_ticket()
 *
 * This routine will generate an FGL test ticket and send it to the currently
 * selected port.  All of this could have been put in a file, read into the
 * buffer and then sent to the printer.
 **************************************************************************************/
void fgl_test_ticket()
{

	strcpy(Buffer,"<RL><RC300,20><F12>BOCA");
	Port_Write();
	strcpy(Buffer,"<RC300,110><RL><F6><HW1,1><VA7>");	//	codename/version
	Port_Write();
	strcpy(Buffer,"<RC300,160><RL><F6><HW1,1><VA9>");	//	common code.
	Port_Write();
	strcpy(Buffer,"<RC370,740><RU>GHOSTWRITER");
	Port_Write();
	strcpy(Buffer,"<RC48,880><RR><F2>FRIENDLY GHOST LANGUAGE");
	Port_Write();
	strcpy(Buffer,"<RC52,862>PLACE LETTERS ANYWHERE");
	Port_Write();
	strcpy(Buffer,"<RC80,1170><F6><RR>VOID");
	Port_Write();
	strcpy(Buffer,"<RC15,1100><F3>TEST TICKET ONLY");
	Port_Write();
	strcpy(Buffer,"<RC8,1010><X2><NXL10>*GHOST 123*");
	Port_Write();
	strcpy(Buffer,"<RR><RC78,915>  CODE 39  ");
	Port_Write();
	strcpy(Buffer,"<RU><RC320,740><BS17,30>TICKET & LABEL PRINTER");
	Port_Write();
	strcpy(Buffer,"<NR><F3><RC0,300>300 DPI PRINT QUALITY");
	Port_Write();
	strcpy(Buffer,"<NR><RC30,300> Print any bar code");
	Port_Write();
	strcpy(Buffer,"<RC70,300><X2><OL6>^CODE 128^");
	Port_Write();
	strcpy(Buffer,"<RC90,250><RR>CODE 128");
	Port_Write();
	strcpy(Buffer,"<RC96,340><X3><NXP10>*MONKEY*");
	Port_Write();
	strcpy(Buffer,"<RC206,340><X2><NP6>*MONKEY*<F2><HW1,1>");
	Port_Write();
	strcpy(Buffer,"<NR><F1><SP180,640><LO1>");
	Port_Write();
	strcpy(Buffer,"<RC371,385><F2><NR>SW1=          SW2=          SW3=");
	Port_Write();
	strcpy(Buffer,"<RC371,824>LEFT = 1  RIGHT = 0");
	Port_Write();
	strcpy(Buffer,"<RC385,150><F11>Script printing is now standard.");
	Port_Write();
	strcpy(Buffer,"<RC430,150><F9>High density printing is clear and readable");
	Port_Write();
	strcpy(Buffer,"<RC450,150><F2>High density printing is clear and readable");
	Port_Write();
	strcpy(Buffer,"<RC470,150><F1>Legal size printing");
	Port_Write();
	strcpy(Buffer,"<p>");	//	Form Feed.
	Port_Write();
}

/**************************************************************************************
 * Routine Name: hp_test_ticket()
 *
 * This routine will generate an HP test ticket.
 **************************************************************************************/
void hp_test_ticket()
{
	FILE *fp;
	//char hp_buffer[PACKET_SIZE+1];	// char string buffer
	long filesize;


	Clear_Buffer();
	
	fp = fopen("hp_id.prn","rb");
	if(fp == NULL)
		printf("\nUnable to find hp_id.prn");
	else
	{
		filesize = filelength(fileno(fp));
		while(filesize > 0)
		{
			if(filesize > SER_PACKET_SIZE)
				i = SER_PACKET_SIZE;
			else
				i = filesize;
			fread(pBuffer, sizeof( char ), i, fp);
			if (i > 0)
				Write_Com = WriteFile(hPort, pBuffer, i, &Bytes_Written, NULL);
			//Port_Write();
			filesize = filesize - i;
		}

		fclose(fp);
	}

}


/**************************************************************************************
 * Routine Name: Perform_Test
 *
 * This routine performs the write tests.
 **************************************************************************************/
void Perform_Test()
{
	switch (w_function)
	{
		case 1:
			Port_Write();			//write an individual command to printer.
		break;

		case 2:
			Port_Write();			//write an ASCII value/command to printer.
		break;
		
		case 3:
			fgl_test_ticket();		//send an entire FGL test ticket to printer.
		break;

		case 4:
			hp_test_ticket();		//send an entire HP test ticket to printer.
		break;

		default:
		break;
	}
	return;
}


/**************************************************************************************
 * Routine Name: Open_Port(int Port_Num)
 *
 * This routine opens a serial port for reading or writing.
 **************************************************************************************/
void Open_Port(int Port_Num)
{

	if (hPort == NULL) 
	{
		//printf("\nOpen_Port for RW Opening = %s \n",port_name);
		hPort = CreateFile (port_name,	// Pointer to the name of the port
		GENERIC_READ | GENERIC_WRITE,	// Access (read-write) mode
		0,								// Share mode
		NULL,							// Pointer to the security attribute
		OPEN_EXISTING,					// How to open the port
		0,								// Port attributes
		NULL);							// Handle to port with attribute

		if (hPort == INVALID_HANDLE_VALUE) 
		{ 
			printf("Could not open %s\n",port_name);   // process error 
		}
		else
		{
			//printf("Setting up port %s\n",port_name); 
			Setup_Port();
		}
		Last_Port_Num = Port_Num;
	}
	else printf("\nhport already in use by %s \n",port_name);

return;
}
/**************************************************************************************
 * Routine Name: Shutdown_Port
 *
 * This routine closes a port with a serial or parallel port printer.
 **************************************************************************************/
void Shutdown_Port()
{

	//If a previous serial or parallel port is open and we've changed ports, 
	//close the first one so the new one can be opened.
	if (hPort != NULL)
	{
		//printf("\nClosing port = %s \n",port_name);
		CloseHandle(hPort);
		hPort = NULL;
		Last_Port_Num = 0;
	}
	return;
}

/**************************************************************************************
 * Routine Name: Setup_Port()
 *
 * This routine will setup the serial or parallel port currently selected
 **************************************************************************************/
void Setup_Port()
{

	//Check the returned handle for INVALID_HANDLE_VALUE and then set the buffer sizes.

	Port_Ready = SetupComm(hPort, PACKET_SIZE, PACKET_SIZE); // set buffer sizes
	
	//Port settings are specified in a Data Communication Block (DCB). 
	//The easiest way to initialize a DCB is to call GetCommState to 
	//fill in its default values, override the values that you want to
	//change and then call SetCommState to set the values.

	//Set up transmission information
	//Also see DOS Mode command:  Mode COM1: 9600,N,8,1
	Port_Ready = GetCommState(hPort, &dcb);

	default_BaudRate = dcb.BaudRate;
	default_Parity = dcb.Parity;
	default_ByteSize = dcb.ByteSize;
	default_StopBits = dcb.StopBits;
	dcb.BaudRate = 9600;
	dcb.Parity = NOPARITY;
	dcb.ByteSize = 8;
	dcb.StopBits = 1;
	dcb.fAbortOnError = TRUE;

	Port_Ready = SetCommState(hPort, &dcb);

	//Communication timeouts are optional but can be set similarly to DCB values
	//This will prevent the program from hanging should the port does not respond.
	Port_Ready = GetCommTimeouts (hPort, &CommTimeouts);

	CommTimeouts.ReadIntervalTimeout = 50;
	CommTimeouts.ReadTotalTimeoutConstant = 50;
	CommTimeouts.ReadTotalTimeoutMultiplier = 10;
	CommTimeouts.WriteTotalTimeoutConstant = 50;
	CommTimeouts.WriteTotalTimeoutMultiplier = 10;

	Port_Ready = SetCommTimeouts (hPort, &CommTimeouts);

	//If all of these API's were successful then the port is ready for use.

	return;
}

/**************************************************************************************
 * Routine Name: Clear_Comm()
 *
 * This routine will clear com port error condition.
 **************************************************************************************/
void Clear_Comm()
{
    COMSTAT  ComStat;
    DWORD    dwErrorFlags;
    DWORD    rc = GetLastError();
    if( rc == ERROR_IO_PENDING ) ClearCommError( hPort, &dwErrorFlags, &ComStat );
	return;
}